#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include "huffman.h"

#if __GNUC__
# define NORETURN __attribute__ ((__noreturn__))
# define PRINTF_STYLE(f, a) __attribute__ ((__format__ (__printf__, f, a)))
# define UNUSED __attribute__ ((__unused__))
#else
# define NORETURN
# define PRINTF_STYLE
# define UNUSED
#endif

static void
PRINTF_STYLE(1, 2)
NORETURN
die(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  abort();
}


#define ASSERT(expr) do {			\
  if (!(expr))					\
    die ("Assertion failed: %s at %s:%d\n",	\
	 #expr, __FILE__, __LINE__);		\
} while (0)

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

static void
one_test (unsigned n, const unsigned char *code_length,
	  unsigned data_length, const int *data,
	  unsigned compressed_length, const char *bits)
{
  struct huffman_node *tree = alloca(HUFFMAN_SIZE(n) * sizeof(*tree));
  unsigned i;
  unsigned j;

  if (!huffman_tree(tree, n, code_length))
    die("huffman_tree failed\n");

  for (i = j = 0; i < data_length; i++)
    {
      struct huffman_ctx ctx;
      unsigned value;
      int res;

      huffman_init (&ctx, tree);
      do
	{
	  ASSERT(j < compressed_length);
	  res = huffman_decode_bit(&ctx, bits[j++], &value);
	}
      while (!res);

      ASSERT (res == 1);
      ASSERT (value == data[i]);
    }
  ASSERT (j == compressed_length);
}

int
main (int argc, char **argv)
{
  {
    /* Too many codes of length 3. */
    const unsigned char clen[] =
      { 2, 3, 3, 4, 3, 3, 3, 4, 3 };

    struct huffman_node tree[HUFFMAN_SIZE(ARRAY_SIZE(clen))];
    if (huffman_tree(tree, ARRAY_SIZE(clen), clen))
      die("huffman_tree accepted invalid coding.\n");
  }
  {
    /* Too many codes. */
    const unsigned char clen[] =
      { 1, 2, 3, 3, 4, 3, 3, 3, 4 };

    struct huffman_node tree[HUFFMAN_SIZE(ARRAY_SIZE(clen))];
    if (huffman_tree(tree, ARRAY_SIZE(clen), clen))
      die("huffman_tree accepted invalid coding.\n");
  }
  {
    /* Too many codes. */
    const unsigned char clen[] =
      { 2, 3, 3, 4, 3, 3, 3, 4, 4 };

    struct huffman_node tree[HUFFMAN_SIZE(ARRAY_SIZE(clen))];
    if (huffman_tree(tree, ARRAY_SIZE(clen), clen))
      die("huffman_tree accepted invalid coding.\n");
  }
  {
    const char code[] =
      {
	0,0,
	0,1,0,	
	0,1,1,
	1,1,1,0,
	1,0,0,
	1,0,1,
	1,1,0,
	1,1,1,1,
	0,0,
      };
    const unsigned char clen[] =
      { 2, 3, 0, 3, 4, 3, 3, 3, 4 };

    const int data[] =
      { 0, 1, 3, 4, 5, 6, 7, 8, 0 };

    one_test (ARRAY_SIZE(clen), clen,
	      ARRAY_SIZE(data), data,
	      ARRAY_SIZE(code), code);
  }
  {
    const char code[] =
      {
	0,
	1, 0, 0, 0
      };
    const unsigned char clen[] =
      { 1, 4 };

    const int data[] =
      { 0, 1 };

    one_test (ARRAY_SIZE(clen), clen,
	      ARRAY_SIZE(data), data,
	      ARRAY_SIZE(code), code);    
  }
  {
    const char code[] =
      {
	0,
	1, 0, 0, 0
      };
    const unsigned char clen[] =
      { 4, 1 };

    const int data[] =
      { 1, 0 };

    one_test (ARRAY_SIZE(clen), clen,
	      ARRAY_SIZE(data), data,
	      ARRAY_SIZE(code), code);    
  }
  return EXIT_SUCCESS;
}
